# Generated by Django 4.2.7 on 2023-11-22 14:46

from django.db import migrations
from pgvector.django import VectorExtension


class Migration(migrations.Migration):
    operations = [VectorExtension()]
